#include <stdio.h>
#include "id.h"
#include "string.h"
#include <sys/types.h>
#include <netinet/in.h>		/* for ntohs(), ntohl() */

extern char *MyName;

/* 
 * Opens an ID file 'fname' with the specified fopen permissions 'perms'.
 * It returns a pointer to the file, and stores the header in idhp.
 * Returns NULL (and closes the file) if the open fails, or if the file
 * doesn't have the right magic number or if the file is not the right
 * version of ID.
 */
FILE *
idOpen(fname, perms, idhp)
char *fname;
char *perms;
struct idhead *idhp;
{
	FILE *fp;
	int exists = 0;

	if (*perms == '+') {	/* ok if file doesn't exist, don't complain */
		perms++;
		exists = 1;
	}
	if ((fp = fopen(fname, perms)) == NULL) {
		if (exists == 0)
			filerr("open", fname);
		return NULL;
	}
	if (fseek(fp, 0L, 0) != 0 || 
	    fread((char *) idhp, sizeof(struct idhead), 1, fp) != 1) {
		filerr("read", fname);
		(void) fclose(fp);
		return NULL;
	}
	if (!strnequ(idhp->idh_magic, IDH_MAGIC, sizeof(idhp->idh_magic))) {
		fprintf(stderr, "%s: Not an id file: `%s'\n", MyName, fname);
		(void) fclose(fp);
		return NULL;
	}
	/* now try to figure out the version number */
	if (idhp->idh_vers != IDH_VERS) {
		if (idhp->idh_vers == htons(IDH_VERS)) {
			/* oops, we're on a different byte order machine */
			/* !! we assume we have long == int */
			idhp->idh_vers = ntohs(idhp->idh_vers);
			idhp->idh_argc = ntohl(idhp->idh_argc);
			idhp->idh_pthc = ntohl(idhp->idh_pthc);
			idhp->idh_namc = ntohl(idhp->idh_namc);
			idhp->idh_vecc = ntohl(idhp->idh_vecc);
			idhp->idh_bsiz = ntohl(idhp->idh_bsiz);
			idhp->idh_argo = ntohl(idhp->idh_argo);
			idhp->idh_namo = ntohl(idhp->idh_namo);
			idhp->idh_endo = ntohl(idhp->idh_endo);
		} else {
			fprintf(stderr, "%s: ID version mismatch (%ld,%ld)\n",
				MyName, idhp->idh_vers, IDH_VERS);
			(void) fclose(fp);
			return NULL;
		}
	}
	if (idhp->idh_argc > MAXIMUM_ARGS) {
		fprintf(stderr, "%s: cannot handle more than %d files\n",
			MyName, MAXIMUM_ARGS);
		exit(1);
	}
			
	return fp;
}

idhton(idhp)
struct idhead *idhp;
{
	/* !! we assume we have long == int */
	idhp->idh_vers = htons(idhp->idh_vers);
	idhp->idh_argc = htonl(idhp->idh_argc);
	idhp->idh_pthc = htonl(idhp->idh_pthc);
	idhp->idh_namc = htonl(idhp->idh_namc);
	idhp->idh_vecc = htonl(idhp->idh_vecc);
	idhp->idh_bsiz = htonl(idhp->idh_bsiz);
	idhp->idh_argo = htonl(idhp->idh_argo);
	idhp->idh_namo = htonl(idhp->idh_namo);
	idhp->idh_endo = htonl(idhp->idh_endo);
}
