/* scanner for text for the 'id' identifier database package. */
/* Mark Moraes, University of Toronto */
/* exports char *getTextId() and void setTextArgs() */
/* 
 * This scanner returns tokens from an input stream considered to
 * consist of strings of characters separated by strings of punctuation.
 * The tokens returned are the strings of characters.
 */
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "bool.h"
#include "id.h"

extern char *bsearch();

/* should get this from sys/types */
#define u_char	unsigned char

#define PC	0x0001	/* punctuation character */
#define	EF	0x0200	/* EOF */

/* !! darn limits. fix this with malloc/realloc. */
#define MAXSTR	512

/*
 * by default, we set all characters for which ispunct() or iswhite() are
 * true to PC with the exception of '-' to rule out hyphens.
 */
static u_char idctype[] = {
	EF,

	/*      0       1       2       3       4       5       6       7   */
	/*    -----   -----   -----   -----   -----   -----   -----   ----- */

	/*000*/	0,	0,	0,	0,	0,	0,	0,	0,
	/*010*/	0,	PC,	PC,	PC,	PC,	PC,	0,	0,
	/*020*/	0,	0,	0,	0,	0,	0,	0,	0,
	/*030*/	0,	0,	0,	0,	0,	0,	0,	0,
	/*040*/	PC,	PC,	PC,	PC,	PC,	PC,	PC,	PC,
	/*050*/	PC,	PC,	PC,	PC,	PC,	0,	PC,	PC,
	/*060*/	0,	0,	0,	0,	0,	0,	0,	0,
	/*070*/	0,	0,	PC,	PC,	PC,	PC,	PC,	PC,
	/*100*/	PC,	0,	0,	0,	0,	0,	0,	0,
	/*110*/	0,	0,	0,	0,	0,	0,	0,	0,
	/*120*/	0,	0,	0,	0,	0,	0,	0,	0,
	/*130*/	0,	0,	0,	PC,	PC,	PC,	PC,	PC,
	/*140*/	PC,	0,	0,	0,	0,	0,	0,	0,
	/*150*/	0,	0,	0,	0,	0,	0,	0,	0,
	/*160*/	0,	0,	0,	0,	0,	0,	0,	0,
	/*170*/	0,	0,	0,	PC,	PC,	PC,	PC,	0,

	/*200*/	0,	0,	0,	0,	0,	0,	0,	0,
	/*210*/	0,	0,	0,	0,	0,	0,	0,	0,
	/*220*/	0,	0,	0,	0,	0,	0,	0,	0,
	/*230*/	0,	0,	0,	0,	0,	0,	0,	0,
	/*240*/	0,	0,	0,	0,	0,	0,	0,	0,
	/*250*/	0,	0,	0,	0,	0,	0,	0,	0,
	/*260*/	0,	0,	0,	0,	0,	0,	0,	0,
	/*270*/	0,	0,	0,	0,	0,	0,	0,	0,
	/*300*/	0,	0,	0,	0,	0,	0,	0,	0,
	/*310*/	0,	0,	0,	0,	0,	0,	0,	0,
	/*320*/	0,	0,	0,	0,	0,	0,	0,	0,
	/*330*/	0,	0,	0,	0,	0,	0,	0,	0,
	/*340*/	0,	0,	0,	0,	0,	0,	0,	0,
	/*350*/	0,	0,	0,	0,	0,	0,	0,	0,
	/*360*/	0,	0,	0,	0,	0,	0,	0,	0,
	/*370*/	0,	0,	0,	0,	0,	0,	0,	0,
};
static u_char *idcp = &idctype[1];

/*
 * Words not to bother indexing on. (it also ignores single characters)
 * These are common words, and would unnecessarily bloat the index. Feel
 * free to add to or delete from this list. Note that I've avoided a few
 * common words that I can think of being used in uncommon contexts
 * which one might want to search for. Sigh. These words are rejected if
 * they, or a capitalized version of them are seen. Inserting very long
 * words will slow down the scanner since it will search this table only
 * if the identifier is shorter than the word with the maximum length in
 * this table. Which is why we have omitted some five character words.
 * Hashing would be faster, perhaps?
 */
static char *throwaway[] = {
	"all", "also", "an", "and", "are", "as", "at",
	"be", "but", "by",
	"can",
	"did",
	"for", "from",
	"have", "has", "how",
	"in", "into", "is", "it",
	"know",
	"like",
	"many", "may", "mine", "my",
	"no", "none", "not",
	"of", "off", "on", "one", "only", "or",
	"that", "the", "them", "then", "they", "this", "to",
	"so",
	"up", "use", "used",
	"was", "we", "were", "what", "when", "why",
	"with",
	"you", "your",
	/* last element must be null */
	0,
};
static int nthrow;

static int
alpha(p1, p2)
char *p1;
char *p2;
{
	return strcmp(*(char **)p1, *(char **)p2);
}

static int
ncmp(s1, s2)
char *s1;
char *s2;
{
    char *p = *(char **)s2;
    int i = *s1 - *p;

    return (i == 0) ? strcmp(s1, p) : i;
}

/*
 * grab the next identifier in the text file fp. flagP is set to a
 * combination of flags that indicate the nature of the token -- we
 * always return IDN_NAME (See id.h for the list)
 */
char *
getTextId(fp, flagP)
FILE *fp;
int *flagP;
{
    int c = EOF;
    int i;
    static int maxthrow = 0;
    static int first = 0;
    static char idBuf[MAXSTR+1];
    char *cp;
    static char *maxcp = idBuf + MAXSTR;

    if (maxthrow == 0) {
	int len;

	for(i = 0; throwaway[i] != 0; i++) {
	    len = strlen(throwaway[i]);
	    if (len > maxthrow)
		maxthrow = len;
	}
	nthrow = i;
	qsort((char *) throwaway, nthrow, sizeof(char *), alpha);
#if 0
	for(i = 0; i < nthrow; i++) {
		printf("%d: %s\n", i, throwaway[i]);
	}
#endif
    }

next:
    if (first == 0) {
	/* 
	 * skip punctuation first time into the scanner for a file.
	 * Since we always skip after getting a token, we don't need to
	 * do this again
	 */
	while ((c = getc(fp)) != EOF && idcp[c] != 0)
	    ;
	if (c != EOF) ungetc(c, fp);
	first++;
    }
    if(feof(fp) || ferror(fp)) {
	first = 0;
	return NULL;
    }
    for(cp = idBuf; cp < maxcp; *cp++ = c) {
	c = getc(fp);
	if (c == EOF || idcp[c] != 0) {
	    break;
	}
    }
    if (cp == maxcp) {
#define FIRSTCHARS  32
	char ebuf[FIRSTCHARS+1];
	ebuf[FIRSTCHARS] = '\0';
	fprintf(stderr, "Warning: token \"%s...\" too long. Max. %d chars.\n",
		strncpy(ebuf, idBuf, FIRSTCHARS), MAXSTR);
	/* skip till punctuation */
	while((c = getc(fp)) != EOF && idcp[c] == 0)
	    ;
    }
    /* null terminate buffer */
    *cp = '\0';
    /* skip punctuation */
    while (c != EOF && idcp[c] != 0) {
	c = getc(fp);
    }
    if (c != EOF)
	ungetc(c, fp);
    i = cp - idBuf;
    /* ignore single characters */
    if (i < 2) {
	goto next;
    }
    /* ignore common words */
    /* 
     * for a table with 57 words, on a Sun3, compiled with gcc, bsearch
     * is slightly faster than a fast inline linear search, on average.
     * linear search with the commonly used words at the start of the
     * array would be faster, but less convenient -- people couldn't add
     * their own words etc in future.
     */
    if (i <= maxthrow) {
	char uc = 0;
	char *search;
	
	if (isascii(idBuf[0]) && isupper(idBuf[0])) {
	    uc = idBuf[0];
	    idBuf[0] = tolower(idBuf[0]);
	}
	search = bsearch((char *) idBuf, (char *) throwaway, nthrow,
			 sizeof(char *), ncmp);
	if (uc)
	    idBuf[0] = uc;
	if (search)
	    goto next;
    }
    *flagP = IDN_NAME;
    return idBuf;
}


/* 
 * !! would be nice to add options to add or delete from the set of
 * punctuation characters. would also be nice to take the list of
 * characters to throw away from a file, determine whether or not to
 * throw away single characters etc.
 */
void
setTextArgs(lang, op, arg)
char *lang;
int op;
char *arg;
{
}
